import logging
from django.db.models.signals import post_delete, post_save, pre_save
from django.dispatch import receiver
from netaddr import IPNetwork

from ipam.models import IPAddress
from netbox_ddns.background_tasks import dns_create, dns_delete
from netbox_ddns.models import DNSStatus, ExtraDNSName
from netbox_ddns.utils import normalize_fqdn

logger = logging.getLogger('netbox_ddns')


@receiver(pre_save, sender=IPAddress)
def store_original_ipaddress(instance: IPAddress, **_kwargs):
    instance.before_save = IPAddress.objects.filter(pk=instance.pk).first()


@receiver(post_save, sender=IPAddress)
def trigger_ddns_update(instance: IPAddress, **_kwargs):
    old_address = instance.before_save.address.ip if instance.before_save else None
    old_dns_name = normalize_fqdn(instance.before_save.dns_name) if instance.before_save else ''

    new_address = IPNetwork(instance.address).ip
    new_dns_name = normalize_fqdn(instance.dns_name)

    extra_dns_names = {normalize_fqdn(extra.name): extra for extra in instance.extradnsname_set.all()}

    if new_address != old_address or new_dns_name != old_dns_name:
        status, created = DNSStatus.objects.get_or_create(ip_address=instance)

        if old_address and old_dns_name and old_dns_name not in extra_dns_names:
            delete = dns_delete.delay(
                dns_name=old_dns_name,
                address=old_address,
                status=status,
            )
        else:
            delete = None

        if new_address and new_dns_name:
            dns_create.delay(
                dns_name=new_dns_name,
                address=new_address,
                status=status,
                depends_on=delete,
            )

    if old_address != new_address:
        # This affects extra names
        for dns_name, status in extra_dns_names.items():
            # Don't touch the main dns_name
            if dns_name == old_dns_name or dns_name == new_dns_name:
                continue

            if old_dns_name:
                delete = dns_delete.delay(
                    dns_name=dns_name,
                    address=old_address,
                    reverse=False,
                    status=status,
                )
            else:
                delete = None

            if new_dns_name:
                dns_create.delay(
                    dns_name=dns_name,
                    address=new_address,
                    reverse=False,
                    status=status,
                    depends_on=delete,
                )


@receiver(post_delete, sender=IPAddress)
def trigger_ddns_delete(instance: IPAddress, **_kwargs):
    old_address = instance.address.ip
    old_dns_name = normalize_fqdn(instance.dns_name)

    if old_address and old_dns_name:
        dns_delete.delay(
            address=old_address,
            dns_name=old_dns_name,
        )


@receiver(pre_save, sender=ExtraDNSName)
def store_original_extra(instance: ExtraDNSName, **_kwargs):
    instance.before_save = ExtraDNSName.objects.filter(pk=instance.pk).first()


@receiver(post_save, sender=ExtraDNSName)
def trigger_extra_ddns_update(instance: ExtraDNSName, **_kwargs):
    address = instance.ip_address.address.ip
    old_dns_name = instance.before_save.name if instance.before_save else ''
    new_dns_name = instance.name

    if new_dns_name != old_dns_name:
        if old_dns_name:
            delete = dns_delete.delay(
                dns_name=old_dns_name,
                address=address,
                reverse=False,
                status=instance,
            )
        else:
            delete = None

        dns_create.delay(
            dns_name=new_dns_name,
            address=address,
            status=instance,
            reverse=False,
            depends_on=delete,
        )


@receiver(post_delete, sender=ExtraDNSName)
def trigger_extra_ddns_delete(instance: ExtraDNSName, **_kwargs):
    address = instance.ip_address.address.ip
    old_dns_name = instance.name

    if old_dns_name == normalize_fqdn(instance.ip_address.dns_name):
        return

    dns_delete.delay(
        dns_name=old_dns_name,
        address=address,
        reverse=False,
    )



VERSION = '1.1.4'

try:
    from extras.plugins import PluginConfig
except ImportError:
    # Dummy for when importing outside of netbox
    class PluginConfig:
        pass


class NetBoxDDNSConfig(PluginConfig):
    name = 'netbox_ddns'
    verbose_name = 'Dynamic DNS'
    version = VERSION
    min_version = '3.0.0'
    max_version = '3.0.999'
    author = 'Sander Steffann'
    author_email = 'sander@steffann.nl'
    description = 'Dynamic DNS Connector for NetBox'
    base_url = 'ddns'
    required_settings = []
    default_settings = {}

    def ready(self):
        super().ready()

        from . import signals


config = NetBoxDDNSConfig



# Generated by Django 3.0.5 on 2020-04-19 16:21

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('netbox_ddns', '0005_extradnsname'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='extradnsname',
            name='reverse_action',
        ),
        migrations.RemoveField(
            model_name='extradnsname',
            name='reverse_rcode',
        ),
    ]



# Generated by Django 3.0.5 on 2020-04-15 10:57

from django.db import migrations


# noinspection PyUnusedLocal
def add_trailing_dots(apps, schema_editor):
    update_trailing_dots(apps, trailing_dot='.')


# noinspection PyUnusedLocal
def remove_trailing_dots(apps, schema_editor):
    update_trailing_dots(apps, trailing_dot='')


def update_trailing_dots(apps, trailing_dot):
    server_model = apps.get_model('netbox_ddns', 'Server')
    zone_model = apps.get_model('netbox_ddns', 'Zone')
    reverse_zone_model = apps.get_model('netbox_ddns', 'ReverseZone')

    for server in server_model.objects.all():
        server.tsig_key_name = server.tsig_key_name.rstrip('.') + trailing_dot
        server.save()

    for zone in zone_model.objects.all():
        zone.name = zone.name.rstrip('.') + trailing_dot
        zone.save()

    for reverse_zone in reverse_zone_model.objects.all():
        reverse_zone.name = reverse_zone.name.rstrip('.') + trailing_dot
        reverse_zone.save()


class Migration(migrations.Migration):
    dependencies = [
        ('netbox_ddns', '0003_dnsstatus'),
    ]

    operations = [
        migrations.RunPython(
            code=add_trailing_dots,
            reverse_code=remove_trailing_dots),
    ]



from django.contrib.auth.context_processors import PermWrapper
from django.template.context_processors import csrf

from extras.plugins import PluginTemplateExtension
from . import tables


# noinspection PyAbstractClass
class DNSInfo(PluginTemplateExtension):
    model = 'ipam.ipaddress'

    def buttons(self):
        """
        A button to force DNS re-provisioning
        """
        context = {
            'perms': PermWrapper(self.context['request'].user),
        }
        context.update(csrf(self.context['request']))
        return self.render('netbox_ddns/ipaddress/dns_refresh_button.html', context)

    def left_page(self):
        """
        An info-box with the status of the DNS modifications and records
        """
        extra_dns_name_table = tables.PrefixTable(list(self.context['object'].extradnsname_set.all()), orderable=False)

        return (
                self.render('netbox_ddns/ipaddress/dns_info.html') +
                self.render('netbox_ddns/ipaddress/dns_extra.html', {
                    'perms': PermWrapper(self.context['request'].user),
                    'extra_dns_name_table': extra_dns_name_table,
                })
        )


template_extensions = [DNSInfo]



# Generated by Django 3.0.5 on 2020-04-19 19:15

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('netbox_ddns', '0006_extradns_cname'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='zone',
            options={'ordering': ('name',), 'verbose_name': 'forward zone', 'verbose_name_plural': 'forward zones'},
        ),
    ]



# Generated by Django 3.0.5 on 2020-04-14 22:52

from django.db import migrations, models
import django.db.models.deletion
import ipam.fields
import netbox_ddns.validators


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Server',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False)),
                ('server', models.CharField(max_length=255, validators=[netbox_ddns.validators.HostnameAddressValidator()])),
                ('tsig_key_name', models.CharField(max_length=255, validators=[netbox_ddns.validators.HostnameValidator()])),
                ('tsig_algorithm', models.CharField(max_length=32)),
                ('tsig_key', models.CharField(max_length=512, validators=[netbox_ddns.validators.validate_base64])),
            ],
            options={
                'verbose_name': 'dynamic DNS Server',
                'verbose_name_plural': 'dynamic DNS Servers',
                'ordering': ('server', 'tsig_key_name'),
                'unique_together': {('server', 'tsig_key_name')},
            },
        ),
        migrations.CreateModel(
            name='Zone',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False)),
                ('name', models.CharField(max_length=255, unique=True, validators=[netbox_ddns.validators.HostnameValidator()])),
                ('server', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='netbox_ddns.Server')),
            ],
            options={
                'verbose_name': 'zone',
                'verbose_name_plural': 'zones',
                'ordering': ('name',),
            },
        ),
        migrations.CreateModel(
            name='ReverseZone',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False)),
                ('prefix', ipam.fields.IPNetworkField(unique=True)),
                ('name', models.CharField(blank=True, max_length=255)),
                ('server', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='netbox_ddns.Server')),
            ],
            options={
                'verbose_name': 'reverse zone',
                'verbose_name_plural': 'reverse zones',
                'ordering': ('prefix',),
            },
        ),
    ]
